#    Copyright (C) 2021  Ryan Raba
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
this module will be included in the api
"""

import pandas as pd
import os
import numpy as np


####################################################
## define functions that compute various indicators
## from basic inputs (high, low, open, close)
####################################################
def BarTrend(highs, lows, opens, closes):
    ratios = (closes - opens) / (highs - lows)
    ratios[np.isnan(ratios)] = np.nanmean(ratios)
    return ratios

def RSI(values, period=14):
    deltas = np.diff(values)
    seed = deltas[:period+1]
    up = seed[seed>=0].sum()/period
    down = -seed[seed<0].sum()/period
    if (down == 0): down = -0.01/period
    rs = up/down
    rsi = np.zeros_like(values)
    rsi[:period] = 1. - 1./(1.+rs)
    for ii in range(period, len(values)):
        delta = deltas[ii-1]
        if delta>0:
            upval = delta
            downval = 0.
        else:
            upval = 0.
            downval = -delta
        up = (up*(period-1) + upval)/period
        down = (down*(period-1) + downval)/period
        if (down == 0): down = -0.01/period
        rs = up/down
        rsi[ii] = 1. - 1./(1.+rs)
    #rsi = list((np.array(rsi)*2.0)-1.0)
    return rsi

def Percent(values):
    perc = np.zeros((len(values)), dtype=float)
    perc[1:] = np.diff(values)/values[:-1]
    return perc

def SMA(values,period):
    weigths = np.repeat(1.0, period)/period
    smas = list(np.convolve(values, weigths, 'valid'))
    pad = [0.0 for _ in range(len(values) - len(smas))]
    return pad + smas

def EMA(values, period):
    period = float(period)
    a = np.array(values)
    for ii in range(1,len(values)):
        a = a[ii] * (2.0 / (period + 1.0)) + a[ii-1] * (1.0 - (2.0 / (period + 1.0)))
    return a

def MACD(values, slow=26, fast=12):
    emaslow = EMA(values, slow)
    emafast = EMA(values, fast)
    macd = emafast - emaslow
    signal = EMA(macd, 5)
    return macd, signal

def Stochastic(closes, highs, lows, period=14):
    so = [0.0 for _ in range(len(closes))]
    for ii in range(period,len(closes)):
        rng = max(ii-period, 0)
        [high, low] = [max(highs[rng:ii]), min(lows[rng:ii])]
        if (high-low) > 0:
            so[ii] = (closes[ii]-low)/(high-low)
    so = SMA(so,3)
    return so

def Breakout(highs, lows):
    diffs = list(np.array(highs) - np.array(lows))
    std = np.std(diffs)
    values = np.zeros((len(diffs)), dtype=int)
    for ii in range(1, len(diffs)):
        if diffs[ii] > 2*std:
            values[ii] = 1
            if highs[ii] < highs[ii-1]:
                values[ii] = -1
    return list(values)





#####################################################################
##
##
##
######################################################################
def FetchData(symbols, apikey, start=None, stop=None, path=None, append=True):
    """
    Download symbol data from iex using provided api key, counts against quota

    Parameters
    ----------
    symbols : list of str
        list of ticker symbols to retrieve
    apikey : str
        api token of the iex account to use
    start : str
        start date of historical prices to retrieve. Format is yyyy-mm-dd.
        Default None uses current date
    stop : str
        stop date of historical prices to retrieve. Format is yyyy-mm-dd.
        Default None uses current date
    path : str
        path of folder to place downloaded data. Default None uses current directory
    append : bool
        append new data to existing file or create if missing. Duplicate dates ignored.
        False will overwrite file. Default True
    """
    import time
    import pandas_datareader.data as web

    if start is None: start = time.strftime("%Y-%m-%d")
    if stop is None: stop = time.strftime("%Y-%m-%d")
    if path is None: path = './'
    if not path.endswith('/'): path = path + '/'
    
    sdf = pd.DataFrame([])
    for symbol in symbols:
        print('fetching %s data...' % symbol, end=' ')
        ddf = web.DataReader(symbol, 'iex', start, stop, api_key=apikey)
        print('%s days' % str(len(ddf)))

        if append and os.path.exists(path+symbol+'.csv'):
            print('merging with existing file for %s...' % symbol)
            pdf = pd.read_csv(path+symbol+'.csv').set_index('date')
            new_dates = np.setdiff1d(pdf.index.values, ddf.index.values)
            ddf = pd.concat([ddf, pdf.loc[new_dates]]).sort_index()
            
        ddf.to_csv(path+symbol+'.csv')
        
    return



##############################################################
def LoadData(symbols=None, path=None):
    """
    Load price data from CSV files
    
    Parameters
    ----------
    symbols : list of str
        list of ticker symbol files to load. Files should be in the form of symbol.csv.
        Default None loads all files in provided directory.
    path : str
        path to symbol data files. Default None uses included demonstration data folder location

    Returns
    -------
    pandas.DataFrame
        symbol dataframe
    """
    import pkg_resources

    if path is None:
        path = pkg_resources.resource_filename('stocksml', 'data/')
    if not path.endswith('/'): path = path + '/'

    sdf = None
    if os.path.exists(path):
        if symbols is None:
            symbols = [ff.split('.csv')[0] for ff in os.listdir(path)]
    
        for symbol in symbols:
            if not os.path.exists(path+symbol+'.csv'): continue
            
            pdf = pd.read_csv(path+symbol+'.csv').set_index('date')
            pdf = pdf.rename(columns=dict([(cc, symbol.lower() + '_' + cc) for cc in pdf.columns]))
            sdf = pdf if sdf is None else sdf.merge(pdf, 'inner', left_index=True, right_index=True)

    symbols = list(np.unique([ss.split('_')[0] for ss in sdf.columns]))
    return sdf, symbols


################################################################
def BuildData(sdf):
    """
    Transform price data from symbol dataframe to training feature set

    Parameters
    ----------
    sdf : pandas.DataFrame
        symbol dataframe
    
    Returns
    -------
    pandas.DataFrame
        feature dataframe
    """
    symbols = list(np.unique([ss.split('_')[0] for ss in sdf.columns]))
    cpdf = pd.DataFrame()
    for ii, ss in enumerate(symbols):
        print('building %s data...' % ss.upper())
        pdf = sdf[[cc for cc in sdf.columns if cc.startswith(ss)]]
        
        # build features
        dm = [Percent(pdf[ss+'_high'].values), Percent(pdf[ss+'_low'].values), Percent(pdf[ss+'_open'].values), Percent(pdf[ss+'_close'].values)]
        dm += [BarTrend(pdf[ss+'_high'].values, pdf[ss+'_low'].values, pdf[ss+'_open'].values, pdf[ss+'_close'].values)]
        dm = np.array(dm).transpose()
    
        # mean normalization
        for kk in range(0, dm.shape[1]):
            if np.std(dm[:, kk]) > 0.0:
                dm[:, kk] = (dm[:, kk] - np.mean(dm[:, kk])) / np.ptp(dm[:, kk])

        # make a dataframe out of the features and merge to combined dataframe
        fpdf = pd.DataFrame(dm, index=pdf.index, columns=[ss + str(ff) for ff in range(dm.shape[1])])
        cpdf = cpdf.merge(fpdf, 'right', left_index=True, right_index=True)
        
    return cpdf



################################################################
## add third dimension for rnn/cnn based models
##
################################################################
def Vectorize(fdf, depth=1):
    """
    Create model training data array from feature set

    Parameters
    ----------
    fdf : pandas.DataFrame
        feature dataframe
    depth : int
        depth of time dimension in trading data. Default is 1 day

    Returns
    -------
    numpy.ndarray
        training data array
    """
    dx = np.repeat(fdf.values[:, None, :], depth, axis=1)
    for ii in range(1,depth):
        dx[:,ii,:] = np.vstack((np.zeros((ii, fdf.values.shape[1])), dx[:-ii,ii,:]))
    
    return dx






#    Copyright (C) 2021  Ryan Raba
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
this module will be included in the api
"""

import os
import numpy as np
import random
os.environ['TF_CPP_MIN_LOG_LEVEL']='2'
random.seed(11)
np.random.seed(13)



###################################################
def BuildModel(shape, layers=[('rnn',32),('dnn',64),('dnn',32)], count=2):
    """
    Build a model with the given structure

    Parameters
    ----------
    shape : tuple
        tuple of training data shape
    layers : list of tuples
        list of tuples defining structure of model. Each tuple is (layer, size) where layer can
        be 'dnn', 'cnn', 'lstm', 'rnn', or 'drop'. Default is a 3-layer model with [('rnn',32),('dnn',64),('dnn',32)]
    count : int
        number of models to build. Default is 2
        
    Returns
    -------
    list of keras.Model
        list of keras Models built, compiled and ready for training
    """
    from keras.models import Model, clone_model
    from keras.layers import Input, Dense, SimpleRNN, LSTM, Conv1D, Flatten, Dropout
    
    ins = Input(shape=shape[1:])
    hh = ins
    for ll, layer in enumerate(layers):
        if layer[0] is 'dnn':
            hh = Dense(layer[1], activation='tanh')(hh)
        elif layer[0] is 'rnn':
            hh = SimpleRNN(layer[1], activation='tanh', return_sequences=(ll+1<len(layers)) and (layers[ll+1][0] in ['cnn','lstm','rnn']))(hh)
        elif layer[0] is 'lstm':
            hh = LSTM(layer[1], activation='tanh', return_sequences=(ll+1<len(layers)) and (layers[ll+1][0] in ['cnn','lstm','rnn']))(hh)
        elif layer[0] is 'cnn':
            hh = Conv1D(layer[1], 3, padding='valid', activation='relu')(hh)
            if (ll+1 >= len(layers)) or (layers[ll+1][0] in ['dnn']):
                hh = Flatten()(hh)
        elif layer[0] is 'drop':
            hh = Dropout(layer[1])(hh)
    
    trade = Dense(5, activation='softmax')(hh)
    limit = Dense(1, activation='tanh')(hh)

    model = Model(inputs=ins, outputs=[trade, limit])
    model.compile(loss=['categorical_crossentropy','mse'], optimizer='adam')

    models = [model]
    for mm in range(count-1):
        models += [clone_model(model)]
        models[-1].compile(loss=['categorical_crossentropy', 'mse'], optimizer='adam')

    return models




###################################################
def TrainModel(models, sdf, dx, symbol, days=5, maxiter=1000, notebook=False):
    """
    Train model against provided data

    Parameters
    ----------
    models : list of keras.Model
        list of prebuilt models to train
    sdf : pandas.DataFrame
        symbol dataframe with price information
    dx : numpy.array
        vectorized training data
    symbol : str
        symbol to use for trading strategy
    days : int
        number of days to use for trading strategy. Default is 5
    maxiter : int
        maximum number of training iterations. Default is 1000
    notebook : bool
        configures live plots for running in a Jupyter notebook.  Default is False
    """
    import matplotlib.pyplot as plt
    from stocksml import EvaluateChoices
    if notebook:
        from IPython import display
    
    fig, ax = plt.subplots(2, 2, figsize=(16,8))
    cum_results = np.zeros((maxiter,3))

    for ee in range(maxiter):
        
        # randomly select a week of data
        ss = np.random.randint(10,dx.shape[0]-6)
        dates = list(sdf.index.values[ss:ss + days])
        
        # each model makes a set of trades for the week
        results = np.zeros((len(models)))
        choices = [[] for _ in range(len(models))]
        for mm in range(len(models)):
            preds = models[mm].predict_on_batch(dx[ss:ss+days])
            choices[mm] = [(np.argmax(preds[0][dd]), preds[1][dd][0]) for dd in range(days)]
            
            # evaluate the performance of the trade choices
            # normalize results to the baseline buy/hold strategy
            results[mm], baseline = EvaluateChoices(sdf, symbol, dates, choices[mm])
            results[mm] = results[mm]/baseline
        
        #print(results, baseline)

        # the model earning the most money wins
        # the winner defines the truth data for this week
        # if neither is successful, skip training this week
        winner = np.argmax(results)
        cum_results[ee][0] = results[winner]
        if np.max(results) <= 1.0: continue
        if np.max(np.abs(np.diff(results))) < 0.0025: continue

        truth = [np.zeros((days,5), dtype=int), np.array([ll[1] for ll in choices[winner]]).reshape(-1,1).clip(-0.05,0.05)]
        for dd in range(days):
            truth[0][dd,choices[winner][dd][0]] = 1

        # train losing model with winners truth data if winner made money
        for mm in range(len(models)):
            if mm == winner: continue
            cum_results[ee][1:] = models[mm].train_on_batch(dx[ss:ss+5], truth)[1:]
            #print('updated model %s'%str(mm), cum_results[ee][1:], results)
            
        # update the plots
        for mm in range(2): ax[mm,0].clear(), ax[mm,1].clear()
        for mm in range(len(models)):
            rc = ax[0, 0].scatter(np.arange(days), [cc[0] for cc in choices[mm]])
            rc = ax[1, 0].scatter(np.arange(days), [cc[1] for cc in choices[mm]])
        train_points = np.where(cum_results[:, 1] > 0)[0]
        rc = ax[0, 1].plot(np.arange(ee), cum_results[:ee, 0])
        rc = ax[1, 1].plot(train_points, cum_results[train_points, 1:])
        ax[1, 1].set_yscale('log'), ax[1,0].set_xlabel('Trading Day'), ax[1,1].set_xlabel('Training Iteration')
    
        if notebook:
            display.clear_output(wait=True)
            display.display(plt.gcf())
        else:
            plt.pause(0.05)
    
    if notebook: plt.close()


#########################################################
def Demo(notebook=False):
    """
    Demonstration of how to use this package
    
    Parameters
    ----------
    notebook : bool
        set live plots for running properly in Jupyter notebooks.  Default is False
    """
    import time
    from stocksml import LoadData, BuildData, Vectorize

    # Globals
    # SYMBOL = 'ARNC'
    #SYMBOLS = ['SPY', 'BND']

    # TRAIN_START = '1992-01-01'
    #TRAIN_START = '2017-01-01'
    #TRAIN_END = time.strftime("%Y-%m-%d")

    # retrieve symbol dataframe
    sdf, symbols = LoadData()

    # build to feature dataframe
    fdf = BuildData(sdf)

    # convert to training data array
    dx = Vectorize(fdf, depth=5)

    models = BuildModel(dx.shape, count=2)

    TrainModel(models, sdf, dx, 'SPY', 5, 1000, notebook)



"""
StocksML package

modules for retrieving data, construction and training models, and evaluating against trading strategies
"""
#__init__.py
from .data import *
from .trade import *
from .model import *



#    Copyright (C) 2021  Ryan Raba
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
this module will be included in the api
"""


###################################################
def EvaluateChoices(pdf, symbol, dates, choices):
    """
    Evaluate trading strategy choices
    
    Parameters
    ----------
    pdf : pandas.DataFrame
        symbol dataframe with price information
    symbol : str
        ticker to use for trading and lookup in symbol dataframe
    dates : list of str
        dates corresponding to choices, should match subset of pdf index values
    choices : list of tuples
        tuple of trading decision and limit for each day. decision is an enum of range 0-4 where
        [buy_limit, buy_sell, hold, sell_limit, sell_buy]. limit is the percent over/under open price (range -1 to 1)

    Returns
    -------
    float
        performance of choices as a fraction of initial cash
    """
    cash = 1000
    shares = 0
    for ii in range(len(choices)):
        didx = pdf.index.get_loc(dates[ii])  # index of this date
        pds = pdf.iloc[didx+1]  # execute decision the next day

        # establish baseline of buy and hold for the duration
        if ii == 0:
            baseline = 1000 / pds[symbol.lower()+'_open']
        if ii == len(choices)-1:
            baseline = (baseline * pds[symbol.lower()+'_open'])/1000

        limit = pds[symbol.lower() + '_open'] * (1 + choices[ii][1])

        # buy limit
        if (choices[ii][0] == 0) and (cash > 0) and (pds[symbol.lower() + '_low'] <= limit):
            shares = cash / min(limit, pds[symbol.lower()+'_open'])
            cash = 0

        # buy open, sell limit
        elif (choices[ii][0] == 1) and (cash > 0):
            shares = cash / pds[symbol.lower()+'_open']
            cash = 0
            if pds[symbol.lower() + '_high'] >= limit:
                cash = shares * max(limit, pds[symbol.lower()+'_open'])
                shares = 0

        # sell limit
        elif (choices[ii][0] == 3) and (shares > 0) and (pds[symbol.lower() + '_high'] >= limit):
            cash = shares * max(limit, pds[symbol.lower()+'_open'])
            shares = 0

        # sell open, buy limit
        elif (choices[ii][0] == 4) and (shares > 0):
            cash = shares * pds[symbol.lower() + '_open']
            shares = 0
            if pds[symbol.lower() + '_low'] <= limit:
                shares = cash / min(limit, pds[symbol.lower()+'_open'])
                cash = 0

    result = (cash + shares*pds[symbol.lower()+'_close'])/1000
    
    return result, baseline



from setuptools import setup, find_packages

with open('README.md', "r") as fid:   #encoding='utf-8'
    long_description = fid.read()

setup(
    name='stocksml',
    version='0.0.1',
    description='Stock Prediction using Machine Learning',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Ryan Raba',
    author_email='stocksml@ryanraba.com',
    url='https://github.com/ryanraba/stocksml',
    license='GPL-3.0',
    packages=find_packages(),
    include_package_data=True,
    install_requires=['matplotlib>=3.1.2',
                      'numpy>=1.18.1',
                      'pandas>=0.25.2',
                      'scipy>=1.4.1',
                      'scikit-learn>=0.22.2',
                      'pandas-datareader>=0.9.0',
                      'keras>=2.4.3',
                      'tensorflow>=2.4.1']
)
